#include "std/string.h"
#include "std/stdlib.h"

size_t strlen(const char* __s)
{
    const char* i = __s;
    for (; *i; ++i)
        ;
    return (i - __s);
}

char* strcpy(char* dest, const char* src)
{
    memcpy(dest, src, strlen(src) + 1);
    return dest;
}

char* strncpy(char* dest, const char* src, size_t n)
{
    size_t i;
    for (i = 0; i < n && src[i] != 0; i++) dest[i] = src[i];
    for (; i < n; i++) dest[i] = 0;
    return dest;
}

size_t strlcpy(char* dest, const char* src, size_t size)
{
    size_t len = strlen(src);
    if (size == 0) return len;
    if (len < (size - 1))
    {
        memcpy(dest, src, len);
        dest[len] = 0;
    }
    else
    {
        memcpy(dest, src, size - 1);
        dest[size - 1] = 0;
    }
    return len;
}

int strcmp(const char* a, const char* b)
{
    while (*a && (*a == *b))
    {
        a++;
        b++;
    }
    return *(const unsigned char*)a - *(const unsigned char*)b;
}

int strncmp(const char* a, const char* b, size_t n)
{
    const char* base = a;
    while (*a && (*a == *b) && (size_t)(a - base) < (n - 1))
    {
        a++;
        b++;
    }
    return *(const unsigned char*)a - *(const unsigned char*)b;
}

char* strncat(char* dest, const char* src, size_t n)
{
    size_t dest_len = strlen(dest);
    size_t i;

    for (i = 0; i < n && *(src + i); i++) *(char*)(dest + dest_len + i) = *(const char*)(src + i);

    *(char*)(dest + dest_len + i) = '\0';

    return dest;
}

char* strcat(char* dest, const char* src)
{
    size_t dest_len = strlen(dest);
    size_t i;

    for (i = 0; *(src + i); i++) *(char*)(dest + dest_len + i) = *(const char*)(src + i);

    *(char*)(dest + dest_len + i) = '\0';

    return dest;
}

char* strstr(char* haystack, const char* needle)
{
    size_t needle_size = strlen(needle);
    size_t haystack_size = strlen(haystack);
    while (*haystack)
    {
        if (*haystack == *needle)
        {
            if (needle_size <= haystack_size)
            {
                if (!strncmp(haystack, needle, needle_size)) return haystack;
            }
            else { return NULL; }
        }
        haystack++;
        haystack_size--;
    }
    return NULL;
}

void* memcpy(void* dest, const void* src, size_t n)
{
    for (size_t i = 0; i < n; ++i) { *((char*)dest + i) = *((const char*)src + i); }
    return dest;
}

void* memset(void* buf, int c, size_t n)
{
    for (size_t i = 0; i < n; ++i) { *((char*)buf + i) = (char)c; }
    return buf;
}

int memcmp(const void* a, const void* b, size_t n)
{
    if (!n) return 0;
    const unsigned char* ap = (const unsigned char*)a;
    const unsigned char* bp = (const unsigned char*)b;
    while (--n && *ap == *bp)
    {
        ap++;
        bp++;
    }
    return *ap - *bp;
}

void* memmove(void* dest, const void* src, size_t n)
{
    if (dest == src) return dest;
    if (dest > src)
        for (long i = n - 1; i >= 0; i++) { *((char*)dest + i) = *((const char*)src + i); }
    else
        for (long i = 0; i < (long)n; i++) { *((char*)dest + i) = *((const char*)src + i); }
    return dest;
}

char* strdup(const char* src)
{
    size_t length = strlen(src);
    char* duplicated = (char*)kmalloc(length + 1);
    memcpy(duplicated, src, length + 1);
    return duplicated;
}

char* strrchr(const char* str, int c)
{
    const char* s = str + strlen(str);
    while (s != str && *s != (char)c) s--;
    if (*s == (char)c) return const_cast<char*>(s);
    return NULL;
}